@@ -21,6 +21,7 @@
  */
 #include "config.h"
 
+#include <glib/gi18n.h>
 #include <gtk/gtk.h>
 
 typedef void (*ActionInvokedCb) (GtkWindow* nw, const char* key);
@@ -43,7 +44,8 @@ typedef struct {
 	GtkWidget* pie_countdown;
 
 	gboolean has_arrow;
-	gboolean enable_transparency;
+	gboolean composited;
+	gboolean action_icons;
 
 	int point_x;
 	int point_y;
@@ -61,11 +63,7 @@ typedef struct {
 	GdkPoint* border_points;
 	size_t num_border_points;
 
-#if GTK_CHECK_VERSION(3, 0, 0)
 	cairo_region_t *window_region;
-#else
-	GdkRegion* window_region;
-#endif
 
 	guchar urgency;
 	glong timeout;
@@ -81,6 +79,23 @@ enum {
 	URGENCY_CRITICAL
 };
 
+gboolean theme_check_init(unsigned int major_ver, unsigned int minor_ver,
+			  unsigned int micro_ver);
+void get_theme_info(char **theme_name, char **theme_ver, char **author,
+		    char **homepage);
+GtkWindow* create_notification(UrlClickedCb url_clicked);
+void set_notification_text(GtkWindow *nw, const char *summary,
+			   const char *body);
+void set_notification_icon(GtkWindow *nw, GdkPixbuf *pixbuf);
+void set_notification_arrow(GtkWidget *nw, gboolean visible, int x, int y);
+void add_notification_action(GtkWindow *nw, const char *text, const char *key,
+			     ActionInvokedCb cb);
+void clear_notification_actions(GtkWindow *nw);
+void move_notification(GtkWidget *nw, int x, int y);
+void set_notification_timeout(GtkWindow *nw, glong timeout);
+void set_notification_hints(GtkWindow *nw, GHashTable *hints);
+void notification_tick(GtkWindow *nw, glong remaining);
+
 //#define ENABLE_GRADIENT_LOOK
 
 #ifdef ENABLE_GRADIENT_LOOK
@@ -103,57 +118,81 @@ enum {
 #define BACKGROUND_OPACITY    0.92
 #define BOTTOM_GRADIENT_HEIGHT 30
 
+static void
+get_background_color (GtkStyleContext *context,
+                      GtkStateFlags    state,
+                      GdkRGBA         *color)
+{
+        GdkRGBA *c;
+
+        g_return_if_fail (color != NULL);
+        g_return_if_fail (GTK_IS_STYLE_CONTEXT (context));
+
+        gtk_style_context_get (context, state,
+                               "background-color", &c,
+                               NULL);
+
+        *color = *c;
+        gdk_rgba_free (c);
+}
+
 static void fill_background(GtkWidget* widget, WindowData* windata, cairo_t* cr)
 {
-	GtkStyle* style;
-	GdkColor* background_color;
+    GtkStyleContext *context;
+    GdkRGBA bg;
 
-	GtkAllocation allocation;
+    GtkAllocation allocation;
 
-	gtk_widget_get_allocation(widget, &allocation);
+    gtk_widget_get_allocation(widget, &allocation);
 
-	#ifdef ENABLE_GRADIENT_LOOK
+    #ifdef ENABLE_GRADIENT_LOOK
 
-		cairo_pattern_t *gradient;
-		int              gradient_y;
+        cairo_pattern_t *gradient;
+        int              gradient_y;
 
-		gradient_y = allocation.height - BOTTOM_GRADIENT_HEIGHT;
+        gradient_y = allocation.height - BOTTOM_GRADIENT_HEIGHT;
 
-	#endif
+    #endif
 
-	style = gtk_widget_get_style(widget);
-	background_color = &style->base[GTK_STATE_NORMAL];
+    context = gtk_widget_get_style_context (windata->win);
 
-	if (windata->enable_transparency)
-	{
-		cairo_set_source_rgba(cr, background_color->red / 65535.0, background_color->green / 65535.0, background_color->blue / 65535.0, BACKGROUND_OPACITY);
-	}
-	else
-	{
-		gdk_cairo_set_source_color(cr, background_color);
-	}
+    gtk_style_context_save (context);
+    gtk_style_context_set_state (context, GTK_STATE_FLAG_NORMAL);
 
-		cairo_rectangle(cr, 0, 0, allocation.width, allocation.height);
+    get_background_color (context, GTK_STATE_FLAG_NORMAL, &bg);
 
-	cairo_fill(cr);
+    gtk_style_context_restore (context);
 
-	#ifdef ENABLE_GRADIENT_LOOK
-		/* Add a very subtle gradient to the bottom of the notification */
-		gradient = cairo_pattern_create_linear(0, gradient_y, 0, allocation.height);
-		cairo_pattern_add_color_stop_rgba(gradient, 0, 0, 0, 0, 0);
-		cairo_pattern_add_color_stop_rgba(gradient, 1, 0, 0, 0, 0.15);
-		cairo_rectangle(cr, 0, gradient_y, allocation.width, BOTTOM_GRADIENT_HEIGHT);
+    if (windata->composited)
+    {
+        cairo_set_source_rgba(cr, bg.red, bg.green, bg.blue, BACKGROUND_OPACITY);
+    }
+    else
+    {
+        gdk_cairo_set_source_rgba (cr, &bg);
+    }
 
-		cairo_set_source(cr, gradient);
-		cairo_fill(cr);
-		cairo_pattern_destroy(gradient);
-	#endif
+    cairo_rectangle(cr, 0, 0, allocation.width, allocation.height);
+
+    cairo_fill(cr);
+
+    #ifdef ENABLE_GRADIENT_LOOK
+        /* Add a very subtle gradient to the bottom of the notification */
+        gradient = cairo_pattern_create_linear(0, gradient_y, 0, allocation.height);
+        cairo_pattern_add_color_stop_rgba(gradient, 0, 0, 0, 0, 0);
+        cairo_pattern_add_color_stop_rgba(gradient, 1, 0, 0, 0, 0.15);
+        cairo_rectangle(cr, 0, gradient_y, allocation.width, BOTTOM_GRADIENT_HEIGHT);
+
+        cairo_set_source(cr, gradient);
+        cairo_fill(cr);
+        cairo_pattern_destroy(gradient);
+    #endif
 }
 
 static void draw_stripe(GtkWidget* widget, WindowData* windata, cairo_t* cr)
 {
-	GtkStyle*        style;
-	GdkColor         color;
+	GtkStyleContext* context;
+	GdkRGBA bg;
 	int              stripe_x;
 	int              stripe_y;
 	int              stripe_height;
@@ -162,7 +201,9 @@ static void draw_stripe(GtkWidget* widget, WindowData* windata, cairo_t* cr)
 		double           r, g, b;
 	#endif
 
-	style = gtk_widget_get_style(widget);
+	context = gtk_widget_get_style_context (widget);
+
+	gtk_style_context_save (context);
 
 	GtkAllocation alloc;
 	gtk_widget_get_allocation(windata->main_hbox, &alloc);
@@ -180,19 +221,27 @@ static void draw_stripe(GtkWidget* widget, WindowData* windata, cairo_t* cr)
 	switch (windata->urgency)
 	{
 		case URGENCY_LOW: // LOW
-			color = style->bg[GTK_STATE_NORMAL];
+			gtk_style_context_set_state (context, GTK_STATE_FLAG_NORMAL);
+			gtk_style_context_add_class (context, GTK_STYLE_CLASS_VIEW);
+			get_background_color (context, GTK_STATE_FLAG_NORMAL, &bg);
+			gdk_cairo_set_source_rgba (cr, &bg);
 			break;
 
 		case URGENCY_CRITICAL: // CRITICAL
-			gdk_color_parse("#CC0000", &color);
+			gdk_rgba_parse (&bg, "#CC0000");
 			break;
 
 		case URGENCY_NORMAL: // NORMAL
 		default:
-			color = style->bg[GTK_STATE_SELECTED];
+			gtk_style_context_set_state (context, GTK_STATE_FLAG_SELECTED);
+			gtk_style_context_add_class (context, GTK_STYLE_CLASS_VIEW);
+			get_background_color (context, GTK_STATE_FLAG_SELECTED, &bg);
+			gdk_cairo_set_source_rgba (cr, &bg);
 			break;
 	}
 
+	gtk_style_context_restore (context);
+
 	cairo_rectangle(cr, stripe_x, stripe_y, STRIPE_WIDTH, stripe_height);
 
 	#ifdef ENABLE_GRADIENT_LOOK
@@ -207,7 +256,7 @@ static void draw_stripe(GtkWidget* widget, WindowData* windata, cairo_t* cr)
 		cairo_fill(cr);
 		cairo_pattern_destroy(gradient);
 	#else
-		gdk_cairo_set_source_color(cr, &color);
+		gdk_cairo_set_source_rgba (cr, &bg);
 		cairo_fill(cr);
 	#endif
 }
@@ -426,11 +475,8 @@ static void create_border_with_arrow(GtkWidget* nw, WindowData* windata)
 
 	g_assert(shape_points != NULL);
 
-#if GTK_CHECK_VERSION(3, 0, 0)
-	/* FIXME!!! */
-#else
-	windata->window_region = gdk_region_polygon(shape_points, windata->num_border_points, GDK_EVEN_ODD_RULE);
-#endif
+	/* FIXME won't work with GTK+3, need a replacement */
+	/*windata->window_region = gdk_region_polygon(shape_points, windata->num_border_points, GDK_EVEN_ODD_RULE);*/
 	g_free(shape_points);
 }
 
@@ -441,7 +487,7 @@ static void draw_border(GtkWidget* widget, WindowData *windata, cairo_t* cr)
 
 	if (windata->has_arrow)
 	{
-		int i;
+		size_t i;
 
 		create_border_with_arrow(windata->win, windata);
 
@@ -453,11 +499,8 @@ static void draw_border(GtkWidget* widget, WindowData *windata, cairo_t* cr)
 		}
 
 		cairo_close_path(cr);
-#if GTK_CHECK_VERSION(3, 0, 0)
-		/* FIXME!!! */
-#else
-		gdk_window_shape_combine_region(windata->win->window, windata->window_region, 0, 0);
-#endif
+		/* FIXME window_region is not set up anyway, see previous fixme */
+		/*gdk_window_shape_combine_region (gtk_widget_get_window (windata->win), windata->window_region, 0, 0);*/
 		g_free(windata->border_points);
 		windata->border_points = NULL;
 	}
@@ -469,18 +512,14 @@ static void draw_border(GtkWidget* widget, WindowData *windata, cairo_t* cr)
 	cairo_stroke(cr);
 }
 
-#if GTK_CHECK_VERSION(3, 0, 0)
-static gboolean paint_window(GtkWidget* widget, cairo_t* cr, WindowData* windata)
-#else
-static gboolean paint_window(GtkWidget* widget, GdkEventExpose* event, WindowData* windata)
-#endif
+static void
+paint_window (GtkWidget  *widget,
+	      cairo_t    *cr,
+	      WindowData *windata)
 {
-	cairo_t*         context;
+	cairo_t*         cr2;
 	cairo_surface_t* surface;
 	GtkAllocation    allocation;
-#if !GTK_CHECK_VERSION(3, 0, 0)
-	cairo_t*         cr;
-#endif
 
 	gtk_widget_get_allocation(windata->win, &allocation);
 
@@ -490,31 +529,32 @@ static gboolean paint_window(GtkWidget* widget, GdkEventExpose* event, WindowDat
 			windata->height = allocation.height;
 	}
 
-	context = gdk_cairo_create(gtk_widget_get_window(widget));
-
-	cairo_set_operator(context, CAIRO_OPERATOR_SOURCE);
+	cairo_set_operator(cr, CAIRO_OPERATOR_SOURCE);
 
 	gtk_widget_get_allocation(widget, &allocation);
 
-	surface = cairo_surface_create_similar(cairo_get_target(context), CAIRO_CONTENT_COLOR_ALPHA, allocation.width, allocation.height);
+	surface = cairo_surface_create_similar (cairo_get_target (cr),
+						CAIRO_CONTENT_COLOR_ALPHA,
+						allocation.width,
+						allocation.height);
 
-#if GTK_CHECK_VERSION (3, 0, 0)
-	cairo_set_source_surface(cr, surface, 0, 0);
-#else
-	cr = cairo_create(surface);
-#endif
+	cr2 = cairo_create (surface);
 
-	fill_background(widget, windata, cr);
-	draw_border(widget, windata, cr);
-	draw_stripe(widget, windata, cr);
+	fill_background(widget, windata, cr2);
+	draw_border(widget, windata, cr2);
+	draw_stripe(widget, windata, cr2);
+	cairo_fill (cr2);
+	cairo_destroy (cr2);
 
-#if !GTK_CHECK_VERSION (3, 0, 0)
-	cairo_destroy(cr);
-#endif
-	cairo_set_source_surface(context, surface, 0, 0);
-	cairo_paint(context);
+	cairo_set_source_surface (cr, surface, 0, 0);
+	cairo_paint(cr);
 	cairo_surface_destroy(surface);
-	cairo_destroy(context);
+}
+
+static gboolean
+on_draw (GtkWidget *widget, cairo_t *cr, WindowData *windata)
+{
+	paint_window (widget, cr, windata);
 
 	return FALSE;
 }
@@ -523,11 +563,7 @@ static void destroy_windata(WindowData* windata)
 {
 	if (windata->window_region != NULL)
 	{
-#if GTK_CHECK_VERSION (3, 0, 0)
 		cairo_region_destroy(windata->window_region);
-#else
-		gdk_region_destroy(windata->window_region);
-#endif
 	}
 
 	g_free(windata);
@@ -602,22 +638,15 @@ GtkWindow* create_notification(UrlClickedCb url_clicked)
 {
 	GtkWidget* spacer;
 	GtkWidget* win;
-	GtkWidget* drawbox;
 	GtkWidget* main_vbox;
 	GtkWidget* hbox;
 	GtkWidget* vbox;
 	GtkWidget* close_button;
 	GtkWidget* image;
-	GtkWidget* alignment;
 	AtkObject* atkobj;
-	GtkRcStyle* rcstyle;
 	WindowData* windata;
 
-#if GTK_CHECK_VERSION (3, 0, 0)
 	GdkVisual *visual;
-#else
-	GdkColormap* colormap;
-#endif
 	GdkScreen* screen;
 
 	windata = g_new0(WindowData, 1);
@@ -625,14 +654,14 @@ GtkWindow* create_notification(UrlClickedCb url_clicked)
 	windata->url_clicked = url_clicked;
 
 	win = gtk_window_new(GTK_WINDOW_POPUP);
+	gtk_window_set_resizable(GTK_WINDOW(win), FALSE);
 	windata->win = win;
 
-	windata->enable_transparency = FALSE;
+	windata->composited = FALSE;
 
 
 	screen = gtk_window_get_screen(GTK_WINDOW(win));
 
-#if GTK_CHECK_VERSION (3, 0, 0)
 	visual = gdk_screen_get_rgba_visual(screen);
 
 	if (visual != NULL)
@@ -641,18 +670,9 @@ GtkWindow* create_notification(UrlClickedCb url_clicked)
 
 		if (gdk_screen_is_composited(screen))
 		{
-			windata->enable_transparency = TRUE;
+			windata->composited = TRUE;
 		}
 	}
-#else
-	colormap = gdk_screen_get_rgba_colormap(screen);
-
-	if (colormap != NULL && gdk_screen_is_composited(screen))
-	{
-		gtk_widget_set_colormap(win, colormap);
-		windata->enable_transparency = TRUE;
-	}
-#endif
 
 	gtk_window_set_title(GTK_WINDOW(win), "Notification");
 	gtk_window_set_type_hint(GTK_WINDOW(win), GDK_WINDOW_TYPE_HINT_NOTIFICATION);
@@ -665,33 +685,18 @@ GtkWindow* create_notification(UrlClickedCb url_clicked)
 
 	g_signal_connect(G_OBJECT(win), "configure_event", G_CALLBACK(configure_event_cb), windata);
 
-	/*
-	 * For some reason, there are occasionally graphics glitches when
-	 * repainting the window. Despite filling the window with a background
-	 * color, parts of the other windows on the screen or the shadows around
-	 * notifications will appear on the notification. Somehow, adding this
-	 * eventbox makes that problem just go away. Whatever works for now.
-	 */
-	drawbox = gtk_event_box_new();
-	gtk_widget_show(drawbox);
-	gtk_container_add(GTK_CONTAINER(win), drawbox);
-
-	main_vbox = gtk_vbox_new(FALSE, 0);
+	main_vbox = gtk_box_new(GTK_ORIENTATION_VERTICAL, 0);
 	gtk_widget_show(main_vbox);
-	gtk_container_add(GTK_CONTAINER(drawbox), main_vbox);
+	gtk_container_add (GTK_CONTAINER (win), main_vbox);
 	gtk_container_set_border_width(GTK_CONTAINER(main_vbox), 1);
 
-#if GTK_CHECK_VERSION (3, 0, 0)
-	g_signal_connect(G_OBJECT(main_vbox), "draw", G_CALLBACK(paint_window), windata);
-#else
-	g_signal_connect(G_OBJECT(main_vbox), "expose_event", G_CALLBACK(paint_window), windata);
-#endif
+	g_signal_connect (G_OBJECT (main_vbox), "draw", G_CALLBACK (on_draw), windata);
 
 	windata->top_spacer = gtk_image_new();
 	gtk_box_pack_start(GTK_BOX(main_vbox), windata->top_spacer, FALSE, FALSE, 0);
 	gtk_widget_set_size_request(windata->top_spacer, -1, DEFAULT_ARROW_HEIGHT);
 
-	windata->main_hbox = gtk_hbox_new(FALSE, 0);
+	windata->main_hbox = gtk_box_new(GTK_ORIENTATION_HORIZONTAL, 0);
 	gtk_widget_show(windata->main_hbox);
 	gtk_box_pack_start(GTK_BOX(main_vbox), windata->main_hbox, FALSE, FALSE, 0);
 
@@ -699,12 +704,12 @@ GtkWindow* create_notification(UrlClickedCb url_clicked)
 	gtk_box_pack_start(GTK_BOX(main_vbox), windata->bottom_spacer, FALSE, FALSE, 0);
 	gtk_widget_set_size_request(windata->bottom_spacer, -1, DEFAULT_ARROW_HEIGHT);
 
-	vbox = gtk_vbox_new(FALSE, 6);
+	vbox = gtk_box_new(GTK_ORIENTATION_VERTICAL, 6);
 	gtk_widget_show(vbox);
 	gtk_box_pack_start(GTK_BOX(windata->main_hbox), vbox, TRUE, TRUE, 0);
 	gtk_container_set_border_width(GTK_CONTAINER(vbox), 10);
 
-	hbox = gtk_hbox_new(FALSE, 6);
+	hbox = gtk_box_new(GTK_ORIENTATION_HORIZONTAL, 6);
 	gtk_widget_show(hbox);
 	gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 0);
 
@@ -716,102 +721,108 @@ GtkWindow* create_notification(UrlClickedCb url_clicked)
 	windata->summary_label = gtk_label_new(NULL);
 	gtk_widget_show(windata->summary_label);
 	gtk_box_pack_start(GTK_BOX(hbox), windata->summary_label, TRUE, TRUE, 0);
-	gtk_misc_set_alignment(GTK_MISC(windata->summary_label), 0, 0);
-	gtk_label_set_line_wrap(GTK_LABEL(windata->summary_label), TRUE);
-#if GTK_CHECK_VERSION (3, 0, 0)
-	gtk_label_set_line_wrap_mode (GTK_LABEL (windata->summary_label), PANGO_WRAP_WORD_CHAR);
+#if GTK_CHECK_VERSION (3, 16, 0)
+	gtk_label_set_xalign (GTK_LABEL (windata->summary_label), 0.0);
+	gtk_label_set_yalign (GTK_LABEL (windata->summary_label), 0.0);
+#else
+	gtk_misc_set_alignment(GTK_MISC(windata->summary_label), 0.0, 0.0);
 #endif
+	gtk_label_set_line_wrap(GTK_LABEL(windata->summary_label), TRUE);
+	gtk_label_set_line_wrap_mode (GTK_LABEL (windata->summary_label), PANGO_WRAP_WORD_CHAR);
 
 	atkobj = gtk_widget_get_accessible(windata->summary_label);
-	atk_object_set_description(atkobj, "Notification summary text.");
+	atk_object_set_description (atkobj, _("Notification summary text."));
 
 	/* Add the close button */
-	alignment = gtk_alignment_new(1, 0, 0, 0);
-	gtk_widget_show(alignment);
-	gtk_box_pack_start(GTK_BOX(hbox), alignment, FALSE, FALSE, 0);
-
 	close_button = gtk_button_new();
 	windata->close_button = close_button;
+	gtk_widget_set_halign (close_button, GTK_ALIGN_END);
+	gtk_widget_set_valign (close_button, GTK_ALIGN_START);
 	gtk_widget_show(close_button);
-	gtk_container_add(GTK_CONTAINER(alignment), close_button);
+	gtk_box_pack_start(GTK_BOX(hbox), close_button, FALSE, FALSE, 0);
 	gtk_button_set_relief(GTK_BUTTON(close_button), GTK_RELIEF_NONE);
 	gtk_container_set_border_width(GTK_CONTAINER(close_button), 0);
 	//gtk_widget_set_size_request(close_button, 20, 20);
-	g_signal_connect_swapped(G_OBJECT(close_button), "clicked", G_CALLBACK(g_object_unref), win);
-
-	rcstyle = gtk_rc_style_new();
-	rcstyle->xthickness = rcstyle->ythickness = 0;
-	gtk_widget_modify_style(close_button, rcstyle);
-	g_object_unref(rcstyle);
+	g_signal_connect_swapped(G_OBJECT(close_button), "clicked", G_CALLBACK(gtk_widget_destroy), win);
 
 	atkobj = gtk_widget_get_accessible(close_button);
-	atk_action_set_description(ATK_ACTION(atkobj), 0, "Closes the notification.");
+	atk_action_set_description(ATK_ACTION(atkobj), 0,
+                               _("Closes the notification."));
 	atk_object_set_name(atkobj, "");
-	atk_object_set_description(atkobj, "Closes the notification.");
+	atk_object_set_description (atkobj, _("Closes the notification."));
 
-	image = gtk_image_new_from_stock(GTK_STOCK_CLOSE, GTK_ICON_SIZE_MENU);
+	image = gtk_image_new_from_icon_name ("window-close", GTK_ICON_SIZE_MENU);
 	gtk_widget_show(image);
 	gtk_container_add(GTK_CONTAINER(close_button), image);
 
-	windata->content_hbox = gtk_hbox_new(FALSE, 6);
+	windata->content_hbox = gtk_box_new(GTK_ORIENTATION_HORIZONTAL, 6);
 	gtk_box_pack_start(GTK_BOX(vbox), windata->content_hbox, FALSE, FALSE, 0);
 
-	windata->iconbox = gtk_hbox_new(FALSE, 0);
+	windata->iconbox = gtk_box_new(GTK_ORIENTATION_HORIZONTAL, 0);
 	gtk_widget_show(windata->iconbox);
 	gtk_box_pack_start(GTK_BOX(windata->content_hbox), windata->iconbox, FALSE, FALSE, 0);
 	gtk_widget_set_size_request(windata->iconbox, BODY_X_OFFSET, -1);
 
 	windata->icon = gtk_image_new();
 	gtk_box_pack_start(GTK_BOX(windata->iconbox), windata->icon, TRUE, TRUE, 0);
-	gtk_misc_set_alignment(GTK_MISC(windata->icon), 0.5, 0.0);
+	gtk_widget_set_halign (windata->icon, GTK_ALIGN_CENTER);
+	gtk_widget_set_valign (windata->icon, GTK_ALIGN_START);
 
-	vbox = gtk_vbox_new(FALSE, 6);
+	vbox = gtk_box_new(GTK_ORIENTATION_VERTICAL, 6);
 	gtk_widget_show(vbox);
 	gtk_box_pack_start(GTK_BOX(windata->content_hbox), vbox, TRUE, TRUE, 0);
 
 	windata->body_label = gtk_label_new(NULL);
+	gtk_widget_show(windata->body_label);
 	gtk_box_pack_start(GTK_BOX(vbox), windata->body_label, TRUE, TRUE, 0);
-	gtk_misc_set_alignment(GTK_MISC(windata->body_label), 0, 0);
-	gtk_label_set_line_wrap(GTK_LABEL(windata->body_label), TRUE);
-#if GTK_CHECK_VERSION (3, 0, 0)
-	gtk_label_set_line_wrap_mode (GTK_LABEL (windata->body_label), PANGO_WRAP_WORD_CHAR);
+#if GTK_CHECK_VERSION (3, 16, 0)
+	gtk_label_set_xalign (GTK_LABEL (windata->body_label), 0.0);
+	gtk_label_set_yalign (GTK_LABEL (windata->body_label), 0.0);
+#else
+	gtk_misc_set_alignment(GTK_MISC(windata->body_label), 0.0, 0.0);
 #endif
+	gtk_label_set_line_wrap(GTK_LABEL(windata->body_label), TRUE);
+	gtk_label_set_line_wrap_mode (GTK_LABEL (windata->body_label), PANGO_WRAP_WORD_CHAR);
+    gtk_label_set_max_width_chars (GTK_LABEL (windata->body_label), 50);
 	g_signal_connect(G_OBJECT(windata->body_label), "activate-link", G_CALLBACK(activate_link), windata);
 
 	atkobj = gtk_widget_get_accessible(windata->body_label);
-	atk_object_set_description(atkobj, "Notification body text.");
+	atk_object_set_description (atkobj, _("Notification body text."));
 
-	alignment = gtk_alignment_new(1, 0.5, 0, 0);
-	gtk_widget_show(alignment);
-	gtk_box_pack_start(GTK_BOX(vbox), alignment, FALSE, TRUE, 0);
-
-	windata->actions_box = gtk_hbox_new(FALSE, 6);
-	gtk_container_add(GTK_CONTAINER(alignment), windata->actions_box);
+	windata->actions_box = gtk_box_new(GTK_ORIENTATION_HORIZONTAL, 6);
+	gtk_widget_set_halign (windata->actions_box, GTK_ALIGN_END);
+	gtk_widget_show(windata->actions_box);
+	gtk_box_pack_start(GTK_BOX(vbox), windata->actions_box, FALSE, TRUE, 0);
 
 	return GTK_WINDOW(win);
 }
 
-void set_notification_hints(GtkWindow* nw, GHashTable* hints)
+void set_notification_hints(GtkWindow *nw, GHashTable *hints)
 {
-	WindowData* windata = g_object_get_data(G_OBJECT(nw), "windata");
+	WindowData *windata = g_object_get_data(G_OBJECT(nw), "windata");
+	GValue *value = NULL, *icon_value = NULL;
 
 	g_assert(windata != NULL);
 
-	GValue* value = (GValue*) g_hash_table_lookup(hints, "urgency");
+	value = (GValue *)g_hash_table_lookup(hints, "urgency");
+	icon_value = (GValue *)g_hash_table_lookup(hints, "action-icons");
 
 	if (value != NULL && G_VALUE_HOLDS_UCHAR(value))
 	{
 		windata->urgency = g_value_get_uchar(value);
 
-		if (windata->urgency == URGENCY_CRITICAL)
-		{
+		if (windata->urgency == URGENCY_CRITICAL) {
 			gtk_window_set_title(GTK_WINDOW(nw), "Critical Notification");
-		}
-		else
-		{
+		} else {
 			gtk_window_set_title(GTK_WINDOW(nw), "Notification");
 		}
 	}
+
+	/* Determine if action-icons have been requested */
+	if (icon_value != NULL && G_VALUE_HOLDS_BOOLEAN(icon_value))
+	{
+		windata->action_icons = g_value_get_boolean(icon_value);
+	}
 }
 
 void set_notification_timeout(GtkWindow* nw, glong timeout)
@@ -852,7 +863,13 @@ void set_notification_text(GtkWindow* nw, const char* summary, const char* body)
 	gtk_label_set_markup(GTK_LABEL(windata->summary_label), str);
 	g_free(str);
 
-	gtk_label_set_markup(GTK_LABEL(windata->body_label), body);
+	if (pango_parse_markup (body, -1, 0, NULL, NULL, NULL, NULL))
+	{
+		gtk_label_set_markup (GTK_LABEL (windata->body_label), body);
+	}
+	else {
+		gtk_label_set_text (GTK_LABEL (windata->body_label), body);
+	}
 
 	if (body == NULL || *body == '\0')
 	{
@@ -867,13 +884,13 @@ void set_notification_text(GtkWindow* nw, const char* summary, const char* body)
 
 	if (body != NULL && *body != '\0')
 	{
-		gtk_widget_size_request(windata->iconbox, &req);
+		gtk_widget_get_preferred_size (windata->iconbox, NULL, &req);
 		/* -1: border width for
 		 * -6: spacing for hbox */
 		gtk_widget_set_size_request(windata->body_label, WIDTH - (1 * 2) - (10 * 2) - req.width - 6, -1);
 	}
 
-	gtk_widget_size_request(windata->close_button, &req);
+		gtk_widget_get_preferred_size (windata->close_button, NULL, &req);
 	/* -1: main_vbox border width
 	 * -10: vbox border width
 	 * -6: spacing for hbox */
@@ -917,54 +934,65 @@ void set_notification_arrow(GtkWidget* nw, gboolean visible, int x, int y)
 	update_spacers(nw);
 }
 
-#if GTK_CHECK_VERSION (3, 0, 0)
-static gboolean countdown_draw_cb(GtkWidget* pie, cairo_t* cr, WindowData* windata)
-#else
-static gboolean countdown_expose_cb(GtkWidget* pie, GdkEventExpose* event, WindowData* windata)
-#endif
+static void
+paint_countdown (GtkWidget  *pie,
+                 cairo_t    *cr,
+                 WindowData *windata)
 {
-	GtkStyle* style;
-	cairo_t* context;
-	cairo_surface_t* surface;
-#if !GTK_CHECK_VERSION(3, 0, 0)
-	cairo_t* cr;
-#endif
+    GtkStyleContext *context;
+    GdkRGBA bg;
+    GtkAllocation alloc;
+    cairo_t* cr2;
+    cairo_surface_t* surface;
 
-	style = gtk_widget_get_style(windata->win);
-	context = gdk_cairo_create(GTK_WINDOW(gtk_widget_get_window(windata->pie_countdown)));
-	GtkAllocation alloc;
-	gtk_widget_get_allocation(pie, &alloc);
-	cairo_set_operator(context, CAIRO_OPERATOR_SOURCE);
-	surface = cairo_surface_create_similar(cairo_get_target(context), CAIRO_CONTENT_COLOR_ALPHA, alloc.width, alloc.height);
-#if GTK_CHECK_VERSION(3, 0, 0)
-	cairo_set_source_surface(cr, surface, 0, 0);
-#else
-	cr = cairo_create(surface);
-#endif
+    context = gtk_widget_get_style_context (windata->win);
 
-	fill_background(pie, windata, cr);
+    gtk_style_context_save (context);
+    gtk_style_context_set_state (context, GTK_STATE_FLAG_SELECTED);
 
-	if (windata->timeout > 0)
-	{
-		gdouble pct = (gdouble) windata->remaining / (gdouble) windata->timeout;
+    get_background_color (context, GTK_STATE_FLAG_SELECTED, &bg);
 
-		gdk_cairo_set_source_color(cr, &style->bg[GTK_STATE_ACTIVE]);
+    gtk_style_context_restore (context);
 
-		cairo_move_to(cr, PIE_RADIUS, PIE_RADIUS);
-		cairo_arc_negative(cr, PIE_RADIUS, PIE_RADIUS, PIE_RADIUS, -G_PI_2, -(pct * G_PI * 2) - G_PI_2);
-		cairo_line_to(cr, PIE_RADIUS, PIE_RADIUS);
-		cairo_fill(cr);
-	}
+    gtk_widget_get_allocation(pie, &alloc);
+    cairo_set_operator(cr, CAIRO_OPERATOR_SOURCE);
+    surface = cairo_surface_create_similar (cairo_get_target(cr),
+                                            CAIRO_CONTENT_COLOR_ALPHA,
+                                            alloc.width,
+                                            alloc.height);
 
-#if !GTK_CHECK_VERSION (3, 0, 0)
-	cairo_destroy(cr);
-#endif
-	cairo_set_source_surface(context, surface, 0, 0);
-	cairo_paint(context);
-	cairo_surface_destroy(surface);
-	cairo_destroy(context);
+    cr2 = cairo_create (surface);
 
-	return TRUE;
+    fill_background (pie, windata, cr2);
+
+    if (windata->timeout > 0)
+    {
+        gdouble pct = (gdouble) windata->remaining / (gdouble) windata->timeout;
+
+        gdk_cairo_set_source_rgba (cr2, &bg);
+
+        cairo_move_to (cr2, PIE_RADIUS, PIE_RADIUS);
+        cairo_arc_negative (cr2, PIE_RADIUS, PIE_RADIUS, PIE_RADIUS, -G_PI_2, -(pct * G_PI * 2) - G_PI_2);
+        cairo_line_to (cr2, PIE_RADIUS, PIE_RADIUS);
+        cairo_fill (cr2);
+    }
+
+    cairo_destroy(cr2);
+
+    cairo_save (cr);
+    cairo_set_source_surface (cr, surface, 0, 0);
+    cairo_paint (cr);
+    cairo_restore (cr);
+
+    cairo_surface_destroy(surface);
+}
+
+static gboolean
+on_countdown_draw (GtkWidget *widget, cairo_t *cr, WindowData *windata)
+{
+	paint_countdown (widget, cr, windata);
+
+	return FALSE;
 }
 
 static void action_clicked_cb(GtkWidget* w, GdkEventButton* event, ActionInvokedCb action_cb)
@@ -989,33 +1017,32 @@ void add_notification_action(GtkWindow* nw, const char* text, const char* key, A
 
 	g_assert(windata != NULL);
 
-	if (!gtk_widget_get_visible(windata->actions_box))
+	if (gtk_widget_get_visible(windata->actions_box))
 	{
-		GtkWidget* alignment;
-
 		gtk_widget_show(windata->actions_box);
 		update_content_hbox_visibility(windata);
 
-		alignment = gtk_alignment_new(1, 0.5, 0, 0);
-		gtk_widget_show(alignment);
-		gtk_box_pack_end(GTK_BOX(windata->actions_box), alignment, FALSE, TRUE, 0);
+		/* Don't try to re-add a pie_countdown */
+		if (!windata->pie_countdown) {
+			windata->pie_countdown = gtk_drawing_area_new();
+			gtk_widget_set_halign (windata->pie_countdown, GTK_ALIGN_END);
+			gtk_widget_show(windata->pie_countdown);
 
-		windata->pie_countdown = gtk_drawing_area_new();
-		gtk_widget_show(windata->pie_countdown);
-		gtk_container_add(GTK_CONTAINER(alignment), windata->pie_countdown);
-		gtk_widget_set_size_request(windata->pie_countdown, PIE_WIDTH, PIE_HEIGHT);
-#if GTK_CHECK_VERSION (3, 0, 0)
-		g_signal_connect(G_OBJECT(windata->pie_countdown), "draw", G_CALLBACK(countdown_draw_cb), windata);
-#else
-		g_signal_connect(G_OBJECT(windata->pie_countdown), "expose_event", G_CALLBACK(countdown_expose_cb), windata);
-#endif
+			gtk_box_pack_end (GTK_BOX (windata->actions_box), windata->pie_countdown, FALSE, TRUE, 0);
+			gtk_widget_set_size_request(windata->pie_countdown,
+						    PIE_WIDTH, PIE_HEIGHT);
+			g_signal_connect(G_OBJECT(windata->pie_countdown), "draw",
+					 G_CALLBACK(on_countdown_draw), windata);
+		}
+	}
+
+	if (windata->action_icons) {
+		button = gtk_button_new_from_icon_name(key, GTK_ICON_SIZE_BUTTON);
+		goto add_button;
 	}
 
 	button = gtk_button_new();
-	gtk_widget_show(button);
-	gtk_box_pack_start(GTK_BOX(windata->actions_box), button, FALSE, FALSE, 0);
-
-	hbox = gtk_hbox_new(FALSE, 6);
+	hbox = gtk_box_new(GTK_ORIENTATION_HORIZONTAL, 6);
 	gtk_widget_show(hbox);
 	gtk_container_add(GTK_CONTAINER(button), hbox);
 
@@ -1030,20 +1057,32 @@ void add_notification_action(GtkWindow* nw, const char* text, const char* key, A
 		GtkWidget* image = gtk_image_new_from_pixbuf(pixbuf);
 		gtk_widget_show(image);
 		gtk_box_pack_start(GTK_BOX(hbox), image, FALSE, FALSE, 0);
-		gtk_misc_set_alignment(GTK_MISC(image), 0.5, 0.5);
+		gtk_widget_set_halign (image, GTK_ALIGN_CENTER);
+		gtk_widget_set_valign (image, GTK_ALIGN_CENTER);
 	}
 
 	label = gtk_label_new(NULL);
 	gtk_widget_show(label);
 	gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 0);
-	gtk_misc_set_alignment(GTK_MISC(label), 0, 0.5);
+#if GTK_CHECK_VERSION (3, 16, 0)
+	gtk_label_set_xalign (GTK_LABEL (label), 0.0);
+	gtk_label_set_yalign (GTK_LABEL (label), 0.5);
+#else
+	gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.5);
+#endif
 	buf = g_strdup_printf("<small>%s</small>", text);
 	gtk_label_set_markup(GTK_LABEL(label), buf);
 	g_free(buf);
 
+add_button:
+	gtk_widget_show(button);
+	gtk_box_pack_start(GTK_BOX(windata->actions_box), button, FALSE, FALSE, 0);
+
 	g_object_set_data(G_OBJECT(button), "_nw", nw);
 	g_object_set_data_full(G_OBJECT(button), "_action_key", g_strdup(key), g_free);
 	g_signal_connect(G_OBJECT(button), "button-release-event", G_CALLBACK(action_clicked_cb), cb);
+
+	gtk_widget_show_all(windata->actions_box);
 }
 
 void clear_notification_actions(GtkWindow* nw)
